from pathlib import Path
from typing import Dict, List, Type

from mycroft_bus_client import Message
from selenium import webdriver

from ..base import Action
from .handlers.base import Handler
from .handlers.selector import SelectionHandler


class BrowserAction(Action):

    HOME_URL = (Path(__file__).parent / "index.html").absolute().as_uri()

    def __init__(self):

        super().__init__()

        self.driver = self.get_driver()

        self.cleanup()  # Default to the home page

        self.is_noisy = False
        self.logger.info(
            "Browser is ready with the following handlers: %s",
            ", ".join(
                [
                    c.__name__.replace("Handler", "")
                    for c in self.get_handlers()
                ]
            ),
        )

    def get_driver(self) -> webdriver.Chrome:

        self.logger.info("Setting up browser")

        options = webdriver.ChromeOptions()
        options.add_argument("disable-infobars")
        options.add_argument("--start-fullscreen")
        options.add_experimental_option(
            "excludeSwitches", ["enable-automation"]
        )
        options.add_argument(
            "--user-data-dir={}/.config/chromium".format(Path.home())
        )

        return webdriver.Chrome(options=options)

    def get_message_types(self) -> Dict[str, callable]:
        return {
            "skill.majel.browser.open": self.handle_single,
            "skill.majel.browser.open-selector": self.handle_multiple,
            "skill.majel.browser.stop": self.handle_stop,
        }

    @staticmethod
    def get_handlers() -> List[Type[Handler]]:
        return sorted(Handler.__subclasses__(), key=lambda c: c.PRIORITY)

    def get_handler(self, payload: str) -> Handler:
        """
        Using the message contents, ask each handler if it can in fact handle
        it.  If it says it can, return an instance of that handler.
        """
        for handler in self.get_handlers():
            self.logger.info("Checking if %s can handle %s", handler, payload)
            if handler := handler.build_from_payload(self.driver, payload):
                self.logger.info("OK: %s can handle %s", handler, payload)
                return handler

    def handle_single(self, message: Message) -> None:
        self.logger.info(str(message.data))
        handler = self.get_handler(message.data.get("url"))
        self.is_noisy = handler.handle()

    def handle_multiple(self, message: Message) -> None:
        handler = SelectionHandler(self.driver, message.data.get("urls"))
        self.is_noisy = handler.handle()

    def handle_stop(self, *args) -> None:
        if self.is_noisy:
            self.cleanup()

    def cleanup(self) -> None:
        self.is_noisy = False
        self.driver.get(self.HOME_URL)



# We have to import these here so they're loaded at start time and recognised
# as subclasses of Handler.  They shouldn't be imported from here.

from .amazon import AmazonHandler
from .basic import BasicHandler
from .netflix import NetflixHandler
from .selector import SelectionHandler
from .youtube import YoutubeHandler



from .base import Handler


class AmazonHandler(Handler):

    IS_NOISY = True

    @classmethod
    def can_handle(cls, payload: str) -> bool:
        return payload.startswith("https://www.amazon.")

    def get_prepared_url(self) -> str:
        """
        Strip superfluous values from Amazon URLs
        """
        return self.payload[: self.payload.index("?")]

    def post_fetch(self):
        self._click(
            "#dv-action-box > "
            "div > "
            "div > "
            "div > "
            "div.abwJ5F.tFxybk._2LF_6p > "
            "span > "
            "div > "
            "a"
        )



from typing import Dict

from ..logger import Loggable


class Action(Loggable):
    def get_message_types(self) -> Dict[str, callable]:
        return {}

    def passive(self) -> None:
        pass

    def cleanup(self) -> None:
        pass



import time

from selenium import webdriver
from selenium.common.exceptions import WebDriverException

from ....logger import Loggable


class Handler(Loggable):
    """
    The abstract handler for all URLs.  Subclass this to handle different URL
    types.
    """

    MAX_WAIT = 10  # Seconds
    PRIORITY = 0

    IS_NOISY = False

    def __init__(self, driver: webdriver.Chrome, payload: str):
        self.payload = payload
        self.driver = driver

    @classmethod
    def build_from_payload(cls, driver, payload: str):
        if cls.can_handle(payload):
            return cls(driver, payload)

    @classmethod
    def can_handle(cls, payload: str) -> bool:
        """
        Return True if this URL is something you can handle.
        """
        return False

    def handle(self) -> bool:

        self.pre_fetch()

        url = self.get_prepared_url()

        self.logger.info("Getting %s", url)
        self.driver.get(url)

        self.post_fetch()

        return self.IS_NOISY

    def pre_fetch(self):
        """
        Do something before the URL has been requested.  I'm not sure how this
        might be used, but it seemed short-sighted not to include a hook for
        it.
        """

    def get_prepared_url(self) -> str:
        """
        Sometimes a URL needs to be modified before it's requested.  If that's
        the case, return the modified version here.
        """
        return self.payload

    def post_fetch(self):
        """
        Do something after the URL has been requested.  Typically this will be
        some Selenium actions.
        """

    def _click(self, selector: str) -> None:
        """
        A Selenium hack: attempt to find the thing we want to click on, and
        keep trying for MAX_WAIT seconds just in case the page loading is slow.
        """

        self.logger.info("Clicking %s", selector)

        start_time = time.time()

        while True:
            try:
                self.logger.info(f"Looking for {selector}")
                self.driver.find_element_by_css_selector(selector).click()
                return
            except (AssertionError, WebDriverException) as e:
                if time.time() - start_time > self.MAX_WAIT:
                    self.logger.error(
                        "Failed to find the thing you wanted to click"
                    )
                    return
                time.sleep(0.5)



from .base import Handler


class BasicHandler(Handler):
    """
    The default behaviour: just go to the page and call it a day.  For this
    reason, the priority should be lowest among all other handlers
    """

    PRIORITY = 1

    @classmethod
    def can_handle(cls, payload: str) -> bool:
        return True



from .base import Handler


class YoutubeHandler(Handler):

    IS_NOISY = True

    @classmethod
    def can_handle(cls, payload: str) -> bool:
        return payload.startswith("https://www.youtube.com/watch")

    def get_prepared_url(self) -> str:
        return self.payload.replace("youtube", "yout-ube")

    def post_fetch(self):
        self._click(".ytp-large-play-button")



from .base import Handler


class NetflixHandler(Handler):

    IS_NOISY = True

    @classmethod
    def can_handle(cls, payload: str) -> bool:
        return payload.startswith("https://www.netflix.com/title/")

    def get_prepared_url(self) -> str:
        return f"https://www.netflix.com/watch/{self.payload[30:]}"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['majel',
 'majel.actions',
 'majel.actions.browser',
 'majel.actions.browser.handlers']

package_data = \
{'': ['*']}

install_requires = \
['kodi-json>=1.0.0,<2.0.0',
 'mycroft-messagebus-client>=0.8.1,<0.9.0',
 'python-mpv>=0.4.6,<0.5.0',
 'requests>=2.23.0,<3.0.0',
 'selenium>=3.141.0,<4.0.0']

entry_points = \
{'console_scripts': ['majel = majel.command:Majel.run']}

setup_kwargs = {
    'name': 'majel',
    'version': '0.1.0',
    'description': 'A front-end for Mycroft that allows you to do cool things like stream video or surf the web',
    'long_description': None,
    'author': 'Daniel Quinn',
    'author_email': 'code@danielquinn.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/danielquinn/majel',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import json
import os
import subprocess
import time

from typing import Any, Dict

import kodijson

from mpv import MPV
from mycroft_bus_client import Message

from .base import Action


class KodiAction(Action):

    NFS_MOUNT_REPLACE = os.getenv("NFS_MOUNT_REPLACE", "").split(",")

    def __init__(self) -> None:

        super().__init__()

        self.endpoint = os.getenv("KODI_ENDPOINT")
        self.client = kodijson.Kodi(self.endpoint)

        self.is_enabled = bool(self.endpoint)

        self._meta = {}
        self.player = None
        self.setup_player()

    def get_message_types(self) -> Dict[str, callable]:

        if not self.is_enabled:
            return {}

        return {
            "skill.majel.kodi.play": self.play,
            "skill.majel.kodi.stop": self.stop,
        }

    def play(self, info: Message) -> None:

        self.cleanup()

        self.logger.info("Media player started for %s", info)

        self._meta = self._get_details(info.data)

        url = self._meta["file"].replace(*self.NFS_MOUNT_REPLACE)
        position = self._meta["resume"]["position"]

        self.player.play(url)

        while self.player.time_remaining is None:
            self.logger.info("Waiting for playback to begin")
            time.sleep(0.1)

        if not position == 0:
            self.logger.info(f"Seeking to {position}")
            self.player.seek(position, reference="absolute")

        self.setup_next()

    def stop(self, *args) -> None:
        self.logger.info("Stopping")
        self.cleanup()

    def passive(self) -> None:

        if not self.player:
            return

        if self.player.time_remaining:
            return

        # We've reached the end of the episode.  Time to see if there's
        # something next.

        episode_id = self._meta.get("next")
        if episode_id:
            self.logger.info("Next episode id: %s", episode_id)
            self.play(
                Message(
                    "skill.majel.kodi.play", data={"episodeid": episode_id}
                )
            )

        return super().passive()

    def cleanup(self) -> None:

        if not self.player:
            return

        if self.player.stream_pos or self._meta:
            self.save_state()
            self.player.terminate()

        self._meta = {}

        self.player = self.get_player()

    def setup_player(self) -> None:

        if self.NFS_MOUNT_REPLACE == [""]:
            self.is_enabled = False
            self.logger.warning(
                "No NFS replacement set.  Assuming there's no NFS share to "
                "play with"
            )
            return

        if not subprocess.call(["mpv", "--version"]) == 0:
            self.is_enabled = False
            self.logger.warning(
                "MPV does not appear to be installed.  We can't play local "
                "video without it."
            )
            return

        self.player = self.get_player()

    def get_player(self) -> MPV:
        return MPV(
            input_default_bindings=True,
            input_vo_keyboard=True,
            osc=True,
            fs=True,
        )

    def process(self, raw_data: str) -> None:

        self.logger.info("Media reference found: %s", raw_data)

        # An empty file means "kill the video"
        if not raw_data:
            return

        info = json.loads(raw_data)

        if not self.player:
            self.logger.error(
                "No NFS mount configured, so %s can't be played", info
            )
            return

        self.play(info)

    def setup_next(self) -> None:
        """
        Ask Kodi if there's a next episode, and if there is, put it into our
        special kodi.__details dict so we can queue it up when this episode
        finishes.
        """

        show_id = self._meta.get("tvshowid")
        episode_id = self._meta.get("episodeid")
        if not show_id:
            return

        self.logger.info("Attempting to get next episode id")

        episodes = self.client.VideoLibrary.GetEpisodes(
            tvshowid=show_id, properties=["title", "playcount"]
        )
        ids = [_["episodeid"] for _ in episodes["result"]["episodes"]]

        try:
            self._meta["next"] = ids[ids.index(episode_id) + 1]
        except IndexError:
            pass  # I guess we're on the last one

    def save_state(self) -> None:
        """
        If the user is fiddling with the play head, `player.time_pos` can have
        a value and then suddenly be `None`, which will cause the math we do
        here to barf.  To work around this, we capture the value early on and
        test that variable for the rest of this method rather than constantly
        referring to `self.player.time_pos`.
        """

        if not self.player:
            return

        t = self.player.time_pos or 1

        self.logger.info(f"Elapsed time: {t}")

        key = "episodeid"
        fn = self.client.VideoLibrary.SetEpisodeDetails
        if "movieid" in self._meta:
            key = "movieid"
            fn = self.client.VideoLibrary.SetMovieDetails

        position = int(t)

        playcount = 0
        if (self.player.time_remaining or 0) < 120:
            playcount = 1
            position = 0

        self.logger.info(
            "Updating Kodi with position=%s, playcount=%s", t, playcount
        )

        fn(
            resume={
                "position": position,
                "total": self._meta["resume"]["total"],
            },
            playcount=playcount,
            **{key: self._meta[key]},
        )

    def _get_details(self, info: Dict[str, str]) -> Dict[str, Any]:
        """
        Return format:

          {
            "file": "nfs://host/path/to/movies/file.extension",
            "label": "Movie Title",
            "movieid": 123,
            "tvshowid": 123,
            "episodeid": 123
            "resume": {"position": 0.0, "total": 0.0},
          }

        For movies, you get the `movieid` but not `tvshowid` or `episodeid`.
        For episodes, you get `tvshowid` and `episodeid`, but no `movieid`.
        """

        id_name, pk = info.popitem()

        key = "episodedetails"
        fn = self.client.VideoLibrary.GetEpisodeDetails
        properties = ["file", "resume", "tvshowid"]
        if id_name == "movieid":
            key = "moviedetails"
            fn = self.client.VideoLibrary.GetMovieDetails
            properties = ["file", "resume"]

        self.logger.info("%s < %s /w %s", fn, properties, key)

        r = fn(**{id_name: pk}, properties=properties)["result"][key]

        self.logger.info(f"Received '{r}' from Kodi")

        return r



import logging


class Loggable:
    """
    Use this mixin to do logging:
      self.logger.debug("My debugging message")
    """

    __logger = None

    @property
    def logger(self) -> logging.Logger:

        if self.__logger:
            return self.__logger

        logging.basicConfig()

        self.__logger = logging.getLogger(f"majel.{self.__class__.__module__}")
        self.__logger.setLevel(logging.DEBUG)

        return self.logger

    @classmethod
    def get_logger(cls) -> logging.Logger:
        return logging.getLogger(f"majel.{cls.__module__}")



#!/usr/bin/env python3

import time

from pathlib import Path

from mycroft_bus_client import MessageBusClient

from .actions.base import Action
from .logger import Loggable


class Majel(Loggable):

    DOTENV = Path(__file__).parent.parent.parent / ".env"

    def __init__(self) -> None:
        self.client = MessageBusClient()
        self.actions = [a() for a in Action.__subclasses__()]

    def __call__(self, *args, **kwargs):

        for action in self.actions:
            for type_, fn in action.get_message_types().items():
                self.client.on(type_, fn)

        self.client.run_forever()

        try:
            while True:
                for action in self.actions:
                    action.passive()
                time.sleep(0.5)
        except KeyboardInterrupt:
            self.logger.info("Exiting")

    @classmethod
    def run(cls):
        """
        I wish I knew a cleaner way to do this, but this appears to be required
        to play nice with the syntax in pyproject.toml.
        """
        cls()()


if __name__ == "__main__":
    Majel()()



import json
import random
import tempfile

from pathlib import Path

from .base import Handler


class SelectionHandler(Handler):
    """
    When multiple URLs are sent back to allow the user to choose which one.
    """

    SCRATCH = tempfile.NamedTemporaryFile(prefix="majel-", suffix=".html")
    TEMPLATE = Path(__file__).parent.parent / "template.html"

    def __init__(self, driver, payload) -> None:
        super().__init__(driver, payload)
        self.pages = json.loads(payload)

    @classmethod
    def can_handle(cls, payload: str) -> bool:
        """
        When the payload isn't a URL at all, but a JSON blob
        """
        return payload.startswith("[")

    def get_prepared_url(self) -> str:
        return f"file://{self.SCRATCH.name}"

    def pre_fetch(self):

        with open(self.TEMPLATE) as f:
            template = f.read()

        links = ""
        for page in self.pages:
            links += '<a class="{}" href="{}">{}</a>'.format(
                self._get_button_colour(), page["url"], page["title"]
            )

        with open(self.SCRATCH.name, "w") as f:
            f.write(
                template.replace("{{ links }}", links).replace(
                    "{{ columns }}", str(self._get_columns())
                )
            )

    def _get_columns(self):
        page_count = len(self.pages)
        if page_count < 4:
            return page_count
        if page_count < 10:
            return 3
        if page_count < 17:
            return 4
        return 5

    def _get_button_colour(self):
        return "lc2375{:02d}".format(random.randint(1, 8))
